'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var xs = _interopDefault(require('xstream'));
var dropRepeats = _interopDefault(require('xstream/extra/dropRepeats'));
var $$observable = _interopDefault(require('symbol-observable'));
var inferno = require('inferno');
var infernoCreateElement = require('inferno-create-element');

/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */
/* global Reflect, Promise */

var extendStatics = Object.setPrototypeOf ||
    ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
    function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };

function __extends(d, b) {
    extendStatics(d, b);
    function __() { this.constructor = d; }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
}

var __assign = Object.assign || function __assign(t) {
    for (var s, i = 1, n = arguments.length; i < n; i++) {
        s = arguments[i];
        for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p)) t[p] = s[p];
    }
    return t;
};

var PROPS_EFFECT = '@@refract/effect/props';
var toProps = function (props) { return ({
    type: PROPS_EFFECT,
    payload: {
        replace: false,
        props: props
    }
}); };
var asProps = function (props) { return ({
    type: PROPS_EFFECT,
    payload: {
        replace: true,
        props: props
    }
}); };

var MOUNT_EVENT = '@@refract/event/mount';
var UNMOUNT_EVENT = '@@refract/event/unmount';
var DataType;
(function (DataType) {
    DataType["EVENT"] = "event";
    DataType["PROPS"] = "props";
    DataType["CALLBACK"] = "callback";
})(DataType || (DataType = {}));
var isEvent = function (eventName) { return function (data, index) {
    return data.type === DataType.EVENT &&
        data.payload.name === eventName;
}; };
var isProps = function (data) { return data.type === DataType.PROPS; };
var isCallback = function (propName) { return function (data) {
    return data.type === DataType.CALLBACK &&
        data.payload.name === propName;
}; };
var createEventData = function (name, value) { return ({
    type: DataType.EVENT,
    payload: {
        name: name,
        value: value
    }
}); };
var createPropsData = function (props) { return ({
    type: DataType.PROPS,
    payload: props
}); };
var createCallbackData = function (name, args) { return ({
    type: DataType.CALLBACK,
    payload: {
        name: name,
        args: args
    }
}); };
var shallowEquals = function (left, right) {
    return left === right ||
        (Object.keys(left).length === Object.keys(right).length &&
            Object.keys(left).every(function (leftKey) { return left[leftKey] === right[leftKey]; }));
};

var subscribeToSink = function (sink, next, error) {
    return sink.subscribe({
        next: next,
        error: error,
        complete: function () { return void 0; }
    });
};
var createComponent = function (instance, dataObservable, pushEvent) {
    var data = function () { return xs.from(dataObservable); };
    var fromEvent = function (eventName, valueTransformer) {
        return data()
            .filter(isEvent(eventName))
            .map(function (data) {
            var value = data.payload.value;
            return valueTransformer ? valueTransformer(value) : value;
        });
    };
    return {
        mount: data()
            .filter(isEvent(MOUNT_EVENT))
            .mapTo(undefined),
        unmount: data()
            .filter(isEvent(UNMOUNT_EVENT))
            .mapTo(undefined),
        observe: function (propName, valueTransformer) {
            if (propName && typeof instance.props[propName] === 'function') {
                return data()
                    .filter(isCallback(propName))
                    .map(function (data) {
                    var args = data.payload.args;
                    return valueTransformer
                        ? valueTransformer(args)
                        : args[0];
                });
            }
            if (propName) {
                return data()
                    .filter(isProps)
                    .map(function (data) {
                    var prop = data.payload[propName];
                    return valueTransformer ? valueTransformer(prop) : prop;
                })
                    .compose(dropRepeats());
            }
            return data()
                .filter(isProps)
                .map(function (data) { return data.payload; })
                .compose(dropRepeats(shallowEquals));
        },
        fromEvent: fromEvent,
        pushEvent: pushEvent,
        useEvent: function (eventName, seedValue) {
            var events$ = fromEvent(eventName);
            var pushEventValue = pushEvent(eventName);
            return [
                seedValue === undefined
                    ? events$
                    : events$.startWith(seedValue),
                pushEventValue
            ];
        }
    };
};

var configureComponent = function (handler, errorHandler) { return function (aperture, instance, isValidElement, isComponentClass) {
    if (isValidElement === void 0) { isValidElement = function () { return false; }; }
    if (isComponentClass === void 0) { isComponentClass = function () { return false; }; }
    instance.state = {
        renderEffect: false,
        children: null,
        props: {}
    };
    var setState = function (state) {
        if (instance.unmounted) {
            return;
        }
        if (instance.mounted) {
            instance.setState(state);
        }
        else {
            instance.state = __assign({}, instance.state, state);
        }
    };
    var finalHandler = function (initialProps, initialContext) {
        var effectHandler = handler(initialProps, initialContext);
        return function (effect) {
            if (isValidElement(effect)) {
                setState({
                    renderEffect: true,
                    children: effect
                });
            }
            else if (effect && effect.type === PROPS_EFFECT) {
                var payload = effect.payload;
                setState({
                    replace: payload.replace,
                    props: payload.props
                });
            }
            else {
                effectHandler(effect);
            }
        };
    };
    var decoratedProps = {};
    var listeners = [];
    var addListener = function (listener) {
        listeners = listeners.concat(listener);
    };
    var removeListener = function (listener) {
        listeners = listeners.filter(function (l) { return l !== listener; });
    };
    var pushEvent = function (eventName) { return function (val) {
        listeners.forEach(function (listener) {
            listener.next(createEventData(eventName, val));
        });
    }; };
    var decorateProp = function (container, prop, propName) {
        if (propName === 'children' || isComponentClass(prop)) {
            return;
        }
        container[propName] = function () {
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            listeners.forEach(function (listener) {
                listener.next(createCallbackData(propName, args));
            });
            return prop.apply(void 0, args);
        };
    };
    Object.keys(instance.props).forEach(function (propName) {
        if (typeof instance.props[propName] === 'function') {
            decorateProp(decoratedProps, instance.props[propName], propName);
        }
    });
    var dataObservable = (_a = {
            subscribe: function (listener) {
                addListener(listener);
                listener.next(createPropsData(instance.props));
                return { unsubscribe: function () { return removeListener(listener); } };
            }
        },
        _a[$$observable] = function () {
            return this;
        },
        _a);
    var component = createComponent(instance, dataObservable, pushEvent);
    var sinkObservable = aperture(instance.props, instance.context)(component);
    var sinkSubscription = subscribeToSink(sinkObservable, finalHandler(instance.props, instance.context), errorHandler
        ? errorHandler(instance.props, instance.context)
        : undefined);
    instance.reDecorateProps = function (nextProps) {
        Object.keys(nextProps).forEach(function (propName) {
            if (typeof instance.props[propName] === 'function' &&
                nextProps[propName] !== instance.props[propName]) {
                decorateProp(decoratedProps, nextProps[propName], propName);
            }
        });
    };
    instance.pushProps = function (props) {
        listeners.forEach(function (listener) {
            listener.next(createPropsData(props));
        });
    };
    instance.triggerMount = function () {
        pushEvent(MOUNT_EVENT)(undefined);
    };
    instance.triggerUnmount = function () {
        pushEvent(UNMOUNT_EVENT)(undefined);
        sinkSubscription.unsubscribe();
    };
    instance.getChildProps = function () {
        var state = instance.state;
        var stateProps = state.props;
        if (state.replace === true) {
            return __assign({}, stateProps, { pushEvent: pushEvent });
        }
        var additionalProps = __assign({}, decoratedProps, { pushEvent: pushEvent });
        if (state.replace === false) {
            return __assign({}, instance.props, stateProps, additionalProps);
        }
        return __assign({}, instance.props, additionalProps);
    };
    instance.havePropsChanged = function (newProps, newState) {
        var state = instance.state;
        if (state.renderEffect) {
            return state.children !== newState.children;
        }
        var haveStatePropsChanged = !shallowEquals(state.props, newState.props);
        if (newState.replace === true) {
            return haveStatePropsChanged;
        }
        var havePropsChanged = !shallowEquals(instance.props, newProps);
        if (newState.replace === false) {
            return havePropsChanged || haveStatePropsChanged;
        }
        return havePropsChanged;
    };
    var _a;
}; };

var Empty = function () { return null; };
var isValidElement = function (value) {
    return Boolean(value) &&
        typeof value === 'object' &&
        'children' in value &&
        'childFlags' in value &&
        'flags' in value &&
        'parentVNode' in value;
};
var isComponentClass = function (ComponentClass) {
    return Boolean(ComponentClass &&
        ComponentClass.prototype &&
        ComponentClass.prototype.componentDidMount);
};
var withEffects = function (handler, errorHandler) { return function (aperture) { return function (BaseComponent) {
    if (BaseComponent === void 0) { BaseComponent = Empty; }
    return /** @class */ (function (_super) {
        __extends(WithEffects, _super);
        function WithEffects(props, context) {
            var _this = _super.call(this, props, context) || this;
            _this.mounted = false;
            _this.unmounted = false;
            configureComponent(handler, errorHandler)(aperture, _this, isValidElement, isComponentClass);
            return _this;
        }
        WithEffects.prototype.componentDidMount = function () {
            this.mounted = true;
            this.triggerMount();
        };
        WithEffects.prototype.componentWillReceiveProps = function (nextProps) {
            this.reDecorateProps(nextProps);
            this.pushProps(nextProps);
        };
        WithEffects.prototype.shouldComponentUpdate = function (nextProps, nextState) {
            return this.havePropsChanged(nextProps, nextState);
        };
        WithEffects.prototype.componentWillUnmount = function () {
            this.unmounted = true;
            this.triggerUnmount();
        };
        WithEffects.prototype.render = function () {
            if (this.state.children) {
                return this.state.children;
            }
            return infernoCreateElement.createElement(BaseComponent, this.getChildProps());
        };
        return WithEffects;
    }(inferno.Component));
}; }; };

var identity = function (arg) { return arg; };
var compose = function () {
    var fns = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        fns[_i] = arguments[_i];
    }
    if (fns.length === 0) {
        return identity;
    }
    if (fns.length === 1) {
        return fns[0];
    }
    return function (arg) {
        return fns.reduceRight(function (previousResult, fn) { return fn(previousResult); }, arg);
    };
};

exports.withEffects = withEffects;
exports.compose = compose;
exports.asProps = asProps;
exports.toProps = toProps;
exports.PROPS_EFFECT = PROPS_EFFECT;
